@@ -38,6 +38,14 @@
 
 
 static void               xfce_appfinder_model_tree_model_init       (GtkTreeModelIface        *iface);
+static void               xfce_appfinder_model_get_property          (GObject                  *object,
+                                                                      guint                     prop_id,
+                                                                      GValue                   *value,
+                                                                      GParamSpec               *pspec);
+static void               xfce_appfinder_model_set_property          (GObject                  *object,
+                                                                      guint                     prop_id,
+                                                                      const GValue             *value,
+                                                                      GParamSpec               *pspec);
 static void               xfce_appfinder_model_finalize              (GObject                  *object);
 static GtkTreeModelFlags  xfce_appfinder_model_get_flags             (GtkTreeModel             *tree_model);
 static gint               xfce_appfinder_model_get_n_columns         (GtkTreeModel             *tree_model);
@@ -93,29 +101,33 @@ struct _XfceAppfinderModelClass
 
 struct _XfceAppfinderModel
 {
-  GObject              __parent__;
-  gint                 stamp;
+  GObject                __parent__;
+  gint                   stamp;
 
-  GSList              *items;
-  GHashTable          *items_hash;
+  GSList                *items;
+  GHashTable            *items_hash;
 
-  GarconMenu          *menu;
+  GarconMenu            *menu;
+  guint                  menu_changed_idle_id;
 
-  GdkPixbuf           *command_icon_small;
-  GdkPixbuf           *command_icon_large;
-  GarconMenuDirectory *command_category;
+  GdkPixbuf             *command_icon;
+  GdkPixbuf             *command_icon_large;
+  GarconMenuDirectory   *command_category;
 
-  GSList              *categories;
+  GSList                *categories;
+  guint                  categories_changed_idle_id;
 
-  guint                collect_idle_id;
-  GSList              *collect_items;
-  GSList              *collect_categories;
-  GThread             *collect_thread;
-  GCancellable        *collect_cancelled;
+  guint                  collect_idle_id;
+  GSList                *collect_items;
+  GSList                *collect_categories;
+  GThread               *collect_thread;
+  GCancellable          *collect_cancelled;
 
-  GFileMonitor        *history_monitor;
-  GFile               *history_file;
-  guint64              history_mtime;
+  GFileMonitor          *history_monitor;
+  GFile                 *history_file;
+  guint64                history_mtime;
+
+  XfceAppfinderIconSize  icon_size;
 };
 
 typedef struct
@@ -128,7 +140,7 @@ typedef struct
   gchar          *tooltip;
   guint           not_visible : 1;
 
-  GdkPixbuf      *icon_small;
+  GdkPixbuf      *icon;
   GdkPixbuf      *icon_large;
 }
 ModelItem;
@@ -147,6 +159,12 @@ enum
   LAST_SIGNAL
 };
 
+enum
+{
+  PROP_0,
+  PROP_ICON_SIZE
+};
+
 
 
 static guint model_signals[LAST_SIGNAL];
@@ -164,8 +182,18 @@ xfce_appfinder_model_class_init (XfceAppfinderModelClass *klass)
   GObjectClass *gobject_class;
 
   gobject_class = G_OBJECT_CLASS (klass);
+  gobject_class->get_property = xfce_appfinder_model_get_property;
+  gobject_class->set_property = xfce_appfinder_model_set_property;
   gobject_class->finalize = xfce_appfinder_model_finalize;
 
+  g_object_class_install_property (gobject_class,
+                                   PROP_ICON_SIZE,
+                                   g_param_spec_uint ("icon-size", NULL, NULL,
+                                                      XFCE_APPFINDER_ICON_SIZE_SMALLEST,
+                                                      XFCE_APPFINDER_ICON_SIZE_LARGEST,
+                                                      XFCE_APPFINDER_ICON_SIZE_DEFAULT_ITEM,
+                                                      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
   model_signals[CATEGORIES_CHANGED] =
     g_signal_new (g_intern_static_string ("categories-changed"),
                   G_TYPE_FROM_CLASS (gobject_class),
@@ -183,12 +211,15 @@ xfce_appfinder_model_init (XfceAppfinderModel *model)
   /* generate a unique stamp */
   model->stamp = g_random_int ();
   model->items_hash = g_hash_table_new (g_str_hash, g_str_equal);
-  model->command_icon_small = xfce_appfinder_model_load_pixbuf (GTK_STOCK_EXECUTE, ICON_SMALL);
-  model->command_icon_large = xfce_appfinder_model_load_pixbuf (GTK_STOCK_EXECUTE, ICON_LARGE);
+  model->icon_size = XFCE_APPFINDER_ICON_SIZE_DEFAULT_ITEM;
+  model->command_icon = xfce_appfinder_model_load_pixbuf (GTK_STOCK_EXECUTE, model->icon_size);
+  model->command_icon_large = xfce_appfinder_model_load_pixbuf (GTK_STOCK_EXECUTE, XFCE_APPFINDER_ICON_SIZE_48);
   model->command_category = xfce_appfinder_model_get_command_category ();
   model->collect_cancelled = g_cancellable_new ();
 
   model->menu = garcon_menu_new_applications ();
+  appfinder_refcount_debug_add (G_OBJECT (model->menu), "main menu");
+
   model->collect_thread = g_thread_create (xfce_appfinder_model_collect_thread, model, TRUE, NULL);
 }
 
@@ -212,6 +243,57 @@ xfce_appfinder_model_tree_model_init (GtkTreeModelIface *iface)
 }
 
 
+static void
+xfce_appfinder_model_get_property (GObject      *object,
+                                   guint         prop_id,
+                                   GValue       *value,
+                                   GParamSpec   *pspec)
+{
+  XfceAppfinderModel *model = XFCE_APPFINDER_MODEL (object);
+
+  switch (prop_id)
+    {
+    case PROP_ICON_SIZE:
+      g_value_set_uint (value, model->icon_size);
+      break;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+    }
+}
+
+
+
+static void
+xfce_appfinder_model_set_property (GObject      *object,
+                                   guint         prop_id,
+                                   const GValue *value,
+                                   GParamSpec   *pspec)
+{
+  XfceAppfinderModel    *model = XFCE_APPFINDER_MODEL (object);
+  XfceAppfinderIconSize  icon_size;
+
+  switch (prop_id)
+    {
+    case PROP_ICON_SIZE:
+      icon_size = g_value_get_uint (value);
+      if (model->icon_size != icon_size)
+        {
+          model->icon_size = icon_size;
+
+          /* trigger a theme change to reload icons */
+          xfce_appfinder_model_icon_theme_changed (model);
+        }
+      break;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+    }
+}
+
+
 
 static void
 xfce_appfinder_model_finalize (GObject *object)
@@ -228,6 +310,12 @@ xfce_appfinder_model_finalize (GObject *object)
     g_source_remove (model->collect_idle_id);
   g_slist_free (model->collect_categories);
 
+  if (G_UNLIKELY (model->categories_changed_idle_id != 0))
+    g_source_remove (model->categories_changed_idle_id);
+
+  if (G_UNLIKELY (model->menu_changed_idle_id != 0))
+    g_source_remove (model->menu_changed_idle_id);
+
   /* stop history file monitoring */
   xfce_appfinder_model_history_monitor_stop (model);
 
@@ -240,15 +328,15 @@ xfce_appfinder_model_finalize (GObject *object)
   g_slist_foreach (model->items, (GFunc) xfce_appfinder_model_item_free, model);
   g_slist_free (model->items);
 
-  g_slist_foreach (model->collect_categories, (GFunc) xfce_appfinder_model_category_free, NULL);
+  g_slist_foreach (model->collect_categories, (GFunc) g_object_unref, NULL);
   g_slist_free (model->collect_categories);
-  g_slist_foreach (model->categories, (GFunc) xfce_appfinder_model_category_free, NULL);
+  g_slist_foreach (model->categories, (GFunc) g_object_unref, NULL);
   g_slist_free (model->categories);
 
   g_hash_table_destroy (model->items_hash);
 
   g_object_unref (G_OBJECT (model->command_icon_large));
-  g_object_unref (G_OBJECT (model->command_icon_small));
+  g_object_unref (G_OBJECT (model->command_icon));
   g_object_unref (G_OBJECT (model->command_category));
 
   APPFINDER_DEBUG ("model finalized");
@@ -281,12 +369,13 @@ xfce_appfinder_model_get_column_type (GtkTreeModel *tree_model,
   switch (column)
     {
     case XFCE_APPFINDER_MODEL_COLUMN_ABSTRACT:
+    case XFCE_APPFINDER_MODEL_COLUMN_TITLE:
     case XFCE_APPFINDER_MODEL_COLUMN_URI:
     case XFCE_APPFINDER_MODEL_COLUMN_COMMAND:
     case XFCE_APPFINDER_MODEL_COLUMN_TOOLTIP:
       return G_TYPE_STRING;
 
-    case XFCE_APPFINDER_MODEL_COLUMN_ICON_SMALL:
+    case XFCE_APPFINDER_MODEL_COLUMN_ICON:
     case XFCE_APPFINDER_MODEL_COLUMN_ICON_LARGE:
       return GDK_TYPE_PIXBUF;
 
@@ -385,21 +474,29 @@ xfce_appfinder_model_get_value (GtkTreeModel *tree_model,
       g_value_set_static_string (value, item->abstract);
       break;
 
+    case XFCE_APPFINDER_MODEL_COLUMN_TITLE:
+      g_value_init (value, G_TYPE_STRING);
+      if (item->item != NULL)
+        g_value_set_static_string (value, garcon_menu_item_get_name (item->item));
+      else if (item->command != NULL)
+        g_value_set_static_string (value, item->command);
+      break;
+
     case XFCE_APPFINDER_MODEL_COLUMN_COMMAND:
       g_value_init (value, G_TYPE_STRING);
       g_value_set_static_string (value, item->command);
       break;
 
-    case XFCE_APPFINDER_MODEL_COLUMN_ICON_SMALL:
-      if (item->icon_small == NULL
+    case XFCE_APPFINDER_MODEL_COLUMN_ICON:
+      if (item->icon == NULL
           && item->item != NULL)
         {
           name = garcon_menu_item_get_icon_name (item->item);
-          item->icon_small = xfce_appfinder_model_load_pixbuf (name, ICON_SMALL);
+          item->icon = xfce_appfinder_model_load_pixbuf (name, model->icon_size);
         }
 
       g_value_init (value, GDK_TYPE_PIXBUF);
-      g_value_set_object (value, item->icon_small);
+      g_value_set_object (value, item->icon);
       break;
 
     case XFCE_APPFINDER_MODEL_COLUMN_ICON_LARGE:
@@ -407,7 +504,7 @@ xfce_appfinder_model_get_value (GtkTreeModel *tree_model,
           && item->item != NULL)
         {
           name = garcon_menu_item_get_icon_name (item->item);
-          item->icon_large = xfce_appfinder_model_load_pixbuf (name, ICON_LARGE);
+          item->icon_large = xfce_appfinder_model_load_pixbuf (name, XFCE_APPFINDER_ICON_SIZE_48);
         }
 
       g_value_init (value, GDK_TYPE_PIXBUF);
@@ -431,11 +528,17 @@ xfce_appfinder_model_get_value (GtkTreeModel *tree_model,
           cat_str = g_strjoinv ("; ", cat_arr);
           g_free (cat_arr);
 
+          comment = garcon_menu_item_get_comment (item->item);
+          if (comment == NULL)
+            comment = "";
+
           item->tooltip = g_markup_printf_escaped ("<b>%s:</b> %s\n"
+                                                   "<b>%s:</b> %s\n"
                                                    "<b>%s:</b> %s\n"
                                                    "<b>%s:</b> %s\n"
                                                    "<b>%s:</b> %s",
                                                    _("Name"), garcon_menu_item_get_name (item->item),
+                                                   _("Comment"), comment,
                                                    _("Command"), garcon_menu_item_get_command (item->item),
                                                    _("Categories"), cat_str,
                                                    _("Filename"), parse_name);
@@ -550,6 +653,34 @@ xfce_appfinder_model_iter_parent (GtkTreeModel *tree_model,
 
 
 
+static gboolean
+xfce_appfinder_model_categories_changed_idle (gpointer data)
+{
+  XfceAppfinderModel *model = XFCE_APPFINDER_MODEL (data);
+
+  appfinder_return_val_if_fail (XFCE_IS_APPFINDER_MODEL (model), FALSE);
+
+  model->categories_changed_idle_id = 0;
+
+  g_signal_emit (G_OBJECT (model), model_signals[CATEGORIES_CHANGED], 0);
+
+  return FALSE;
+}
+
+
+
+static void
+xfce_appfinder_model_categories_changed (XfceAppfinderModel *model)
+{
+  if (model->categories_changed_idle_id == 0)
+    {
+      model->categories_changed_idle_id =
+        g_idle_add_full (G_PRIORITY_HIGH_IDLE, xfce_appfinder_model_categories_changed_idle, model, NULL);
+    }
+}
+
+
+
 static gboolean
 xfce_appfinder_model_collect_idle (gpointer user_data)
 {
@@ -610,9 +741,9 @@ xfce_appfinder_model_collect_idle (gpointer user_data)
       model->categories = model->collect_categories;
       model->collect_categories = NULL;
 
-      g_signal_emit (G_OBJECT (model), model_signals[CATEGORIES_CHANGED], 0);
+      xfce_appfinder_model_categories_changed (model);
 
-      g_slist_foreach (tmp, (GFunc) xfce_appfinder_model_category_free, NULL);
+      g_slist_foreach (tmp, (GFunc) g_object_unref, NULL);
       g_slist_free (tmp);
     }
 
@@ -709,6 +840,9 @@ xfce_appfinder_model_item_new (GarconMenuItem *menu_item)
   item = g_slice_new0 (ModelItem);
   item->item = g_object_ref (G_OBJECT (menu_item));
 
+  appfinder_refcount_debug_add (G_OBJECT (menu_item),
+     garcon_menu_item_get_desktop_id (menu_item));
+
   command = garcon_menu_item_get_command (menu_item);
   if (G_LIKELY (command != NULL))
     {
@@ -737,6 +871,7 @@ xfce_appfinder_model_item_changed (GarconMenuItem     *menu_item,
   GtkTreeIter  iter;
   GtkTreePath *path;
   GPtrArray   *categories;
+  gboolean     old_not_visible;
 
   /* lookup the item in the list */
   for (li = model->items, idx = 0; li != NULL; li = li->next, idx++)
@@ -749,6 +884,8 @@ xfce_appfinder_model_item_changed (GarconMenuItem     *menu_item,
 
           APPFINDER_DEBUG ("update item %s", garcon_menu_item_get_desktop_id (menu_item));
 
+          old_not_visible = item->not_visible;
+
           xfce_appfinder_model_item_free (item, model);
 
           item = xfce_appfinder_model_item_new (menu_item);
@@ -768,6 +905,10 @@ xfce_appfinder_model_item_changed (GarconMenuItem     *menu_item,
           gtk_tree_model_row_changed (GTK_TREE_MODEL (model), path, &iter);
           gtk_tree_path_free (path);
 
+          /* update the categories if the visibility changed */
+          if (old_not_visible != item->not_visible)
+            xfce_appfinder_model_categories_changed (model);
+
           break;
         }
     }
@@ -790,8 +931,8 @@ xfce_appfinder_model_item_free (gpointer            data,
     }
   if (model != NULL && item->command != NULL)
     g_hash_table_remove (model->items_hash, item->command);
-  if (item->icon_small != NULL)
-    g_object_unref (G_OBJECT (item->icon_small));
+  if (item->icon != NULL)
+    g_object_unref (G_OBJECT (item->icon));
   if (item->icon_large != NULL)
     g_object_unref (G_OBJECT (item->icon_large));
   if (item->categories != NULL)
@@ -879,7 +1020,7 @@ xfce_appfinder_model_history_insert (XfceAppfinderModel *model,
   /* add new command */
   item = g_slice_new0 (ModelItem);
   item->command = g_strdup (command);
-  item->icon_small = g_object_ref (G_OBJECT (model->command_icon_small));
+  item->icon = g_object_ref (G_OBJECT (model->command_icon));
   item->icon_large = g_object_ref (G_OBJECT (model->command_icon_large));
   model->items = g_slist_insert_sorted (model->items, item, xfce_appfinder_model_item_compare);
 
@@ -1016,6 +1157,7 @@ xfce_appfinder_model_history_monitor (XfceAppfinderModel *model,
 
       /* monitor the file for changes */
       model->history_monitor = g_file_monitor_file (file, G_FILE_MONITOR_NONE, model->collect_cancelled, &error);
+      appfinder_refcount_debug_add (G_OBJECT (model->history_monitor), "history file monitor");
       if (model->history_monitor != NULL)
         {
           APPFINDER_DEBUG ("monitor history file %s", path);
@@ -1042,31 +1184,16 @@ static gint
 xfce_appfinder_model_category_compare (gconstpointer a,
                                        gconstpointer b)
 {
-  const CategoryItem *cat_a = a;
-  const CategoryItem *cat_b = b;
+  appfinder_return_val_if_fail (GARCON_IS_MENU_DIRECTORY (a), 0);
+  appfinder_return_val_if_fail (GARCON_IS_MENU_DIRECTORY (b), 0);
 
-  appfinder_return_val_if_fail (GARCON_IS_MENU_DIRECTORY (cat_a->directory), 0);
-  appfinder_return_val_if_fail (GARCON_IS_MENU_DIRECTORY (cat_b->directory), 0);
-
-  return g_utf8_collate (garcon_menu_directory_get_name (cat_a->directory),
-                         garcon_menu_directory_get_name (cat_b->directory));
+  return g_utf8_collate (garcon_menu_directory_get_name (GARCON_MENU_DIRECTORY (a)),
+                         garcon_menu_directory_get_name (GARCON_MENU_DIRECTORY (b)));
 }
 
 
 
-void
-xfce_appfinder_model_category_free (CategoryItem *item)
-{
-  if (item->directory != NULL)
-    g_object_unref (G_OBJECT (item->directory));
-  if (item->pixbuf != NULL)
-    g_object_unref (G_OBJECT (item->pixbuf));
-  g_slice_free (CategoryItem, item);
-}
-
-
-
-static gboolean
+static inline gboolean
 xfce_appfinder_model_ptr_array_find (GPtrArray     *array,
                                      gconstpointer  data)
 {
@@ -1104,7 +1231,7 @@ xfce_appfinder_model_collect_history (XfceAppfinderModel *model,
         {
           item = g_slice_new0 (ModelItem);
           item->command = g_strndup (contents, end - contents);
-          item->icon_small = g_object_ref (G_OBJECT (model->command_icon_small));
+          item->icon = g_object_ref (G_OBJECT (model->command_icon));
           item->icon_large = g_object_ref (G_OBJECT (model->command_icon_large));
           model->collect_items = g_slist_prepend (model->collect_items, item);
         }
@@ -1150,6 +1277,24 @@ xfce_appfinder_model_collect_item (const gchar    *desktop_id,
 
 
 
+static void
+xfce_appfinder_model_directory_changed (GarconMenu *menu)
+{
+  GarconMenu *parent;
+
+  appfinder_return_if_fail (GARCON_IS_MENU (menu));
+
+  while (menu != NULL)
+    {
+      parent = garcon_menu_get_parent (menu);
+      if (parent == NULL)
+        g_signal_emit_by_name (menu, "reload-required");
+      menu = parent;
+    }
+}
+
+
+
 static gboolean
 xfce_appfinder_model_collect_items (GarconMenu           *menu,
                                     GCancellable         *cancelled,
@@ -1161,7 +1306,6 @@ xfce_appfinder_model_collect_items (GarconMenu           *menu,
   GList               *menus, *li;
   GarconMenuDirectory *directory;
   gboolean             has_items = FALSE;
-  CategoryItem        *citem;
   GarconMenuItemPool  *pool;
   CollectContext      *context;
 
@@ -1178,6 +1322,9 @@ xfce_appfinder_model_collect_items (GarconMenu           *menu,
   directory = garcon_menu_get_directory (menu);
   if (directory != NULL)
     {
+      appfinder_refcount_debug_add (G_OBJECT (directory),
+          garcon_menu_directory_get_name (directory));
+
       if (!garcon_menu_directory_get_visible (directory))
         return FALSE;
 
@@ -1186,6 +1333,9 @@ xfce_appfinder_model_collect_items (GarconMenu           *menu,
         category = directory;
     }
 
+  g_signal_connect (G_OBJECT (menu), "directory-changed",
+      G_CALLBACK (xfce_appfinder_model_directory_changed), NULL);
+
   /* collect all the items in the menu's pool. we walk
    * the pool directory to avoid the sorting of items
    * that is done within garcon for _get_elements and
@@ -1215,19 +1365,19 @@ xfce_appfinder_model_collect_items (GarconMenu           *menu,
   menus = garcon_menu_get_menus (menu);
   for (li = menus; li != NULL; li = li->next)
     {
+      appfinder_refcount_debug_add (G_OBJECT (li->data),
+          garcon_menu_element_get_name (li->data));
+
       if (xfce_appfinder_model_collect_items (li->data, cancelled, category, items,
                                               categories, desktop_ids))
-        has_items = TRUE;
+        {
+          has_items = TRUE;
+        }
     }
   g_list_free (menus);
 
-  if (directory != NULL
-      && has_items)
-    {
-      citem = g_slice_new0 (CategoryItem);
-      citem->directory = g_object_ref (G_OBJECT (directory));
-      *categories = g_slist_prepend (*categories, citem);
-    }
+  if (directory != NULL && has_items)
+    *categories = g_slist_prepend (*categories, g_object_ref (G_OBJECT (directory)));
 
   return has_items;
 }
@@ -1273,28 +1423,32 @@ xfce_appfinder_model_menu_changed_remove (gpointer key,
 
 
 
-static void
-xfce_appfinder_model_menu_changed (GarconMenu         *menu,
-                                   XfceAppfinderModel *model)
+static gboolean
+xfce_appfinder_model_menu_changed_idle (gpointer data)
 {
-  GSList       *li, *lp;
-  GHashTable   *old_items;
-  ModelItem    *item;
-  const gchar  *desktop_id;
-  GSList       *collect_items = NULL;
-  GSList       *collect_categories = NULL;
-  GSList       *tmp;
-  guint         idx;
-  GtkTreeIter   iter;
-  GtkTreePath  *path;
-  CategoryItem *cat_a, *cat_b;
+  XfceAppfinderModel *model = XFCE_APPFINDER_MODEL (data);
+  GarconMenu         *menu = model->menu;
+  GSList             *li, *lp;
+  GHashTable         *old_items;
+  ModelItem          *item;
+  ModelItem          *old_item;
+  const gchar        *desktop_id;
+  GSList             *collect_items = NULL;
+  GSList             *collect_categories = NULL;
+  GSList             *tmp;
+  guint               idx;
+  GtkTreeIter         iter;
+  GtkTreePath        *path;
+  GSList             *old_li;
+  GError             *error = NULL;
 
-  appfinder_return_if_fail (GARCON_IS_MENU (menu));
-  appfinder_return_if_fail (XFCE_IS_APPFINDER_MODEL (model));
-  appfinder_return_if_fail (model->menu == menu);
+  appfinder_return_val_if_fail (GARCON_IS_MENU (menu), FALSE);
+  appfinder_return_val_if_fail (XFCE_IS_APPFINDER_MODEL (model), FALSE);
 
   APPFINDER_DEBUG ("menu changed");
 
+  model->menu_changed_idle_id = 0;
+
   /* add all the old items in a garcon database */
   old_items = g_hash_table_new (g_str_hash, g_str_equal);
   for (li = model->items; li != NULL; li = li->next)
@@ -1307,6 +1461,12 @@ xfce_appfinder_model_menu_changed (GarconMenu         *menu,
         }
     }
 
+  if (!garcon_menu_load (model->menu, NULL, &error))
+    {
+      g_warning ("Failed to reload the root menu: %s", error->message);
+      g_error_free (error);
+    }
+
   xfce_appfinder_model_collect_menu (menu, NULL, &collect_items, &collect_categories);
 
   for (li = collect_items; li != NULL; li = li->next)
@@ -1316,8 +1476,30 @@ xfce_appfinder_model_menu_changed (GarconMenu         *menu,
       g_assert (GARCON_IS_MENU_ITEM (item->item));
 
       desktop_id = garcon_menu_item_get_desktop_id (item->item);
-      if (g_hash_table_remove (old_items, desktop_id))
+
+      /* check if desktop id already exists */
+      old_li = g_hash_table_lookup (old_items, desktop_id);
+      if (old_li != NULL)
         {
+          /* check if the file location is still equal */
+          old_item = old_li->data;
+          if (!garcon_menu_element_equal (GARCON_MENU_ELEMENT (old_item->item),
+                                          GARCON_MENU_ELEMENT (item->item)))
+            {
+              APPFINDER_DEBUG ("%s file location changed", desktop_id);
+
+              /* do not remove from the hash table and insert the new item */
+              goto insert;
+            }
+
+          /* remove from the table */
+          g_hash_table_remove (old_items, desktop_id);
+
+          /* steal the categories */
+          g_ptr_array_unref (old_item->categories);
+          old_item->categories = item->categories;
+          item->categories = NULL;
+
           /* this is not interesting, since those items are also updated
            * by the GarconMenuItem::changed signal, so don't touch the
            * item in the model */
@@ -1325,6 +1507,8 @@ xfce_appfinder_model_menu_changed (GarconMenu         *menu,
         }
       else
         {
+          insert:
+
           /* insert new item in the list */
           model->items = g_slist_insert_sorted (model->items, item, xfce_appfinder_model_item_compare);
 
@@ -1344,10 +1528,10 @@ xfce_appfinder_model_menu_changed (GarconMenu         *menu,
 
           /* watch and add the command */
           g_signal_connect (G_OBJECT (item->item), "changed",
-                          G_CALLBACK (xfce_appfinder_model_item_changed), model);
+              G_CALLBACK (xfce_appfinder_model_item_changed), model);
 
           if (G_LIKELY (item->command != NULL))
-             g_hash_table_insert (model->items_hash, item->command, item);
+              g_hash_table_insert (model->items_hash, item->command, item);
         }
     }
 
@@ -1359,27 +1543,32 @@ xfce_appfinder_model_menu_changed (GarconMenu         *menu,
 
   /* update the new categories */
   collect_categories = g_slist_sort (collect_categories, xfce_appfinder_model_category_compare);
-  for (li = collect_categories, lp = model->categories; li != NULL && lp != NULL; li = li->next, lp = lp->next)
-    {
-      cat_a = li->data;
-      cat_b = lp->data;
 
-      /* different items or the list length is different */
-      if (!garcon_menu_directory_equal (cat_a->directory, cat_b->directory)
-          || ((li->next == NULL) != (lp->next == NULL)))
-        {
-          tmp = model->categories;
-          model->categories = collect_categories;
-          collect_categories = tmp;
+  /* swap lists */
+  tmp = model->categories;
+  model->categories = collect_categories;
 
-          g_signal_emit (G_OBJECT (model), model_signals[CATEGORIES_CHANGED], 0);
+  /* always update the categories, because the pointers changed */
+  xfce_appfinder_model_categories_changed (model);
 
-          break;
-        }
-    }
+  g_slist_foreach (tmp, (GFunc) g_object_unref, NULL);
+  g_slist_free (tmp);
 
-  g_slist_foreach (collect_categories, (GFunc) xfce_appfinder_model_category_free, NULL);
-  g_slist_free (collect_categories);
+  return FALSE;
+}
+
+
+
+static void
+xfce_appfinder_model_menu_changed (GarconMenu         *menu,
+                                   XfceAppfinderModel *model)
+{
+  appfinder_return_if_fail (GARCON_IS_MENU (menu));
+  appfinder_return_if_fail (XFCE_IS_APPFINDER_MODEL (model));
+  appfinder_return_if_fail (model->menu == menu);
+
+  if (model->menu_changed_idle_id == 0)
+    model->menu_changed_idle_id = g_idle_add (xfce_appfinder_model_menu_changed_idle, model);
 }
 
 
@@ -1473,6 +1662,7 @@ xfce_appfinder_model_get (void)
     {
       model = g_object_new (XFCE_TYPE_APPFINDER_MODEL, NULL);
       g_object_add_weak_pointer (G_OBJECT (model), (gpointer) &model);
+      appfinder_refcount_debug_add (G_OBJECT (model), "appfinder-model");
       APPFINDER_DEBUG ("allocate new model");
     }
 
@@ -1484,8 +1674,40 @@ xfce_appfinder_model_get (void)
 GSList *
 xfce_appfinder_model_get_categories (XfceAppfinderModel *model)
 {
+  GSList              *categories = NULL;
+  GSList              *li, *lp;
+  gboolean             visible;
+  ModelItem           *item;
+  GarconMenuDirectory *category;
+
   appfinder_return_val_if_fail (XFCE_IS_APPFINDER_MODEL (model), NULL);
-  return model->categories;
+
+  for (li = model->categories; li != NULL; li = li->next)
+    {
+      category = li->data;
+      appfinder_assert (GARCON_IS_MENU_DIRECTORY (category));
+      visible = FALSE;
+
+      /* check if the category has a visible item */
+      for (lp = model->items; lp != NULL; lp = lp->next)
+        {
+          item = lp->data;
+          if (!item->not_visible
+              && xfce_appfinder_model_ptr_array_find (item->categories, category))
+            {
+              visible = TRUE;
+              break;
+            }
+        }
+
+      if (!visible)
+        continue;
+
+      categories = g_slist_prepend (categories, category);
+    }
+
+  /* we return in reversed order here */
+  return categories;
 }
 
 
@@ -1644,13 +1866,25 @@ xfce_appfinder_model_execute (XfceAppfinderModel  *model,
 
 
 GdkPixbuf *
-xfce_appfinder_model_load_pixbuf (const gchar *icon_name,
-                                  gint         size)
+xfce_appfinder_model_load_pixbuf (const gchar           *icon_name,
+                                  XfceAppfinderIconSize  icon_size)
 {
   GdkPixbuf    *pixbuf = NULL;
   GdkPixbuf    *scaled;
   gchar        *p, *name;
   GtkIconTheme *icon_theme;
+  gint          size;
+
+  switch (icon_size)
+    {
+    case XFCE_APPFINDER_ICON_SIZE_SMALLEST: size = 16;  break;
+    case XFCE_APPFINDER_ICON_SIZE_SMALLER:  size = 24;  break;
+    case XFCE_APPFINDER_ICON_SIZE_SMALL:    size = 36;  break;
+    case XFCE_APPFINDER_ICON_SIZE_NORMAL:   size = 48;  break;
+    case XFCE_APPFINDER_ICON_SIZE_LARGE:    size = 64;  break;
+    case XFCE_APPFINDER_ICON_SIZE_LARGER:   size = 96;  break;
+    case XFCE_APPFINDER_ICON_SIZE_LARGEST:  size = 128; break;
+    }
 
   APPFINDER_DEBUG ("load icon %s at %dpx", icon_name, size);
 
@@ -1711,6 +1945,8 @@ xfce_appfinder_model_load_pixbuf (const gchar *icon_name,
       pixbuf = scaled;
     }
 
+  appfinder_refcount_debug_add (G_OBJECT (pixbuf), icon_name);
+
   return pixbuf;
 }
 
@@ -1794,7 +2030,7 @@ xfce_appfinder_model_get_icon_for_command (XfceAppfinderModel *model,
               && item->item != NULL)
             {
               icon_name = garcon_menu_item_get_icon_name (item->item);
-              item->icon_large = xfce_appfinder_model_load_pixbuf (icon_name, ICON_LARGE);
+              item->icon_large = xfce_appfinder_model_load_pixbuf (icon_name, XFCE_APPFINDER_ICON_SIZE_48);
             }
 
           return g_object_ref (G_OBJECT (item->icon_large));
@@ -1818,16 +2054,17 @@ xfce_appfinder_model_icon_theme_changed (XfceAppfinderModel *model)
 
   appfinder_return_if_fail (XFCE_IS_APPFINDER_MODEL (model));
 
-  APPFINDER_DEBUG ("icon theme changed");
+  APPFINDER_DEBUG ("icon theme or size changed, updating %d items",
+                   g_slist_length (model->items));
 
   /* reload the command icons */
-  if (model->command_icon_small != NULL)
-    g_object_unref (G_OBJECT (model->command_icon_small));
-  model->command_icon_small = xfce_appfinder_model_load_pixbuf (GTK_STOCK_EXECUTE, ICON_SMALL);
+  if (model->command_icon != NULL)
+    g_object_unref (G_OBJECT (model->command_icon));
+  model->command_icon = xfce_appfinder_model_load_pixbuf (GTK_STOCK_EXECUTE, model->icon_size);
 
   if (model->command_icon_large != NULL)
     g_object_unref (G_OBJECT (model->command_icon_large));
-  model->command_icon_large = xfce_appfinder_model_load_pixbuf (GTK_STOCK_EXECUTE, ICON_LARGE);
+  model->command_icon_large = xfce_appfinder_model_load_pixbuf (GTK_STOCK_EXECUTE, XFCE_APPFINDER_ICON_SIZE_48);
 
   /* update the model items */
   for (li = model->items, idx = 0; li != NULL; li = li->next, idx++)
@@ -1835,10 +2072,10 @@ xfce_appfinder_model_icon_theme_changed (XfceAppfinderModel *model)
       item = li->data;
       item_changed = FALSE;
 
-      if (item->icon_small != NULL)
+      if (item->icon != NULL)
         {
-          g_object_unref (G_OBJECT (item->icon_small));
-          item->icon_small = NULL;
+          g_object_unref (G_OBJECT (item->icon));
+          item->icon = NULL;
           item_changed = TRUE;
         }
       if (item->icon_large != NULL)
@@ -1850,7 +2087,7 @@ xfce_appfinder_model_icon_theme_changed (XfceAppfinderModel *model)
 
       if (item->item == NULL)
         {
-          item->icon_small = g_object_ref (G_OBJECT (model->command_icon_small));
+          item->icon = g_object_ref (G_OBJECT (model->command_icon));
           item->icon_large = g_object_ref (G_OBJECT (model->command_icon_large));
         }
 
@@ -1900,6 +2137,7 @@ xfce_appfinder_model_get_command_category (void)
                                "name", _("Commands History"),
                                "icon-name", GTK_STOCK_EXECUTE,
                                NULL);
+      appfinder_refcount_debug_add (G_OBJECT (category), "commands");
       g_object_add_weak_pointer (G_OBJECT (category), (gpointer) &category);
     }
 
