@@ -39,6 +39,8 @@
 static void xfce_appfinder_preferences_response          (GtkWidget                *window,
                                                           gint                      response_id,
                                                           XfceAppfinderPreferences *preferences);
+static void xfce_appfinder_preferences_beside_sensitive  (GtkWidget                *show_icons,
+                                                          GtkWidget                *text_beside_icon);
 static void xfce_appfinder_preferences_clear_history     (XfceAppfinderPreferences *preferences);
 static void xfce_appfinder_preferences_action_add        (XfceAppfinderPreferences *preferences);
 static void xfce_appfinder_preferences_action_remove     (GtkWidget                *button,
@@ -96,6 +98,7 @@ xfce_appfinder_preferences_init (XfceAppfinderPreferences *preferences)
 {
   GObject     *object;
   GtkTreePath *path;
+  GObject     *icons;
 
   preferences->channel = xfconf_channel_get ("xfce4-appfinder");
 
@@ -116,6 +119,28 @@ xfce_appfinder_preferences_init (XfceAppfinderPreferences *preferences)
   xfconf_g_property_bind (preferences->channel, "/always-center", G_TYPE_BOOLEAN,
                           G_OBJECT (object), "active");
 
+  icons = gtk_builder_get_object (GTK_BUILDER (preferences), "icon-view");
+  xfconf_g_property_bind (preferences->channel, "/icon-view", G_TYPE_BOOLEAN,
+                          G_OBJECT (icons), "active");
+
+  object = gtk_builder_get_object (GTK_BUILDER (preferences), "text-beside-icons");
+  xfconf_g_property_bind (preferences->channel, "/text-beside-icons", G_TYPE_BOOLEAN,
+                          G_OBJECT (object), "active");
+
+  g_signal_connect (G_OBJECT (icons), "toggled",
+      G_CALLBACK (xfce_appfinder_preferences_beside_sensitive), object);
+  xfce_appfinder_preferences_beside_sensitive (GTK_WIDGET (icons), GTK_WIDGET (object));
+
+  object = gtk_builder_get_object (GTK_BUILDER (preferences), "item-icon-size");
+  gtk_combo_box_set_active (GTK_COMBO_BOX (object), XFCE_APPFINDER_ICON_SIZE_DEFAULT_ITEM);
+  xfconf_g_property_bind (preferences->channel, "/item-icon-size", G_TYPE_UINT,
+                          G_OBJECT (object), "active");
+
+  object = gtk_builder_get_object (GTK_BUILDER (preferences), "category-icon-size");
+  gtk_combo_box_set_active (GTK_COMBO_BOX (object), XFCE_APPFINDER_ICON_SIZE_DEFAULT_CATEGORY);
+  xfconf_g_property_bind (preferences->channel, "/category-icon-size", G_TYPE_UINT,
+                          G_OBJECT (object), "active");
+
   object = gtk_builder_get_object (GTK_BUILDER (preferences), "button-clear");
   g_signal_connect_swapped (G_OBJECT (object), "clicked",
       G_CALLBACK (xfce_appfinder_preferences_clear_history), preferences);
@@ -165,6 +190,16 @@ xfce_appfinder_preferences_response (GtkWidget                *window,
 
 
 
+static void
+xfce_appfinder_preferences_beside_sensitive (GtkWidget *show_icons,
+                                             GtkWidget *text_beside_icon)
+{
+  gtk_widget_set_sensitive (text_beside_icon,
+      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (show_icons)));
+}
+
+
+
 static void
 xfce_appfinder_preferences_clear_history (XfceAppfinderPreferences *preferences)
 {
@@ -392,24 +427,24 @@ xfce_appfinder_preferences_action_populate (XfceAppfinderPreferences *preference
   if (G_LIKELY (array != NULL))
     {
       for (i = 0; i < array->len; i++)
-      {
-        value = g_ptr_array_index (array, i);
-        appfinder_assert (value != NULL);
-        unique_id = g_value_get_int (value);
+        {
+          value = g_ptr_array_index (array, i);
+          appfinder_assert (value != NULL);
+          unique_id = g_value_get_int (value);
 
-        g_snprintf (prop, sizeof (prop), "/actions/action-%d/pattern", unique_id);
-        pattern = xfconf_channel_get_string (preferences->channel, prop, NULL);
+          g_snprintf (prop, sizeof (prop), "/actions/action-%d/pattern", unique_id);
+          pattern = xfconf_channel_get_string (preferences->channel, prop, NULL);
 
-        gtk_list_store_insert_with_values (GTK_LIST_STORE (store), &iter, i,
-                                           COLUMN_UNIQUE_ID, unique_id,
-                                           COLUMN_PATTERN, pattern,
-                                           -1);
+          gtk_list_store_insert_with_values (GTK_LIST_STORE (store), &iter, i,
+                                             COLUMN_UNIQUE_ID, unique_id,
+                                             COLUMN_PATTERN, pattern,
+                                             -1);
 
-        if (restore_id == unique_id)
-          gtk_tree_selection_select_iter (preferences->selection, &iter);
+          if (restore_id == unique_id)
+            gtk_tree_selection_select_iter (preferences->selection, &iter);
 
-        g_free (pattern);
-      }
+          g_free (pattern);
+        }
 
       xfconf_array_free (array);
     }
