@@ -63,6 +63,11 @@ static GSList             *windows = NULL;
 static gboolean            service_owner = FALSE;
 static XfceAppfinderModel *model_cache = NULL;
 
+#ifdef DEBUG
+static GHashTable         *objects_table = NULL;
+static guint               objects_table_count = 0;
+#endif
+
 
 
 static GOptionEntry option_entries[] =
@@ -81,6 +86,68 @@ static void appfinder_dbus_unregister (DBusConnection *dbus_connection);
 
 
 
+#ifdef DEBUG
+static void
+appfinder_refcount_debug_weak_notify (gpointer  data,
+                                      GObject  *where_the_object_was)
+{
+  /* remove the unreffed object pixbuf from the table */
+  if (!g_hash_table_remove (objects_table, where_the_object_was))
+    appfinder_assert_not_reached ();
+}
+
+
+
+static void
+appfinder_refcount_debug_print (gpointer object,
+                                gpointer desc,
+                                gpointer data)
+{
+  APPFINDER_DEBUG ("object %p (type = %s, desc = %s) not released",
+                   object, G_OBJECT_TYPE_NAME (object), (gchar *) desc);
+}
+
+
+
+void
+appfinder_refcount_debug_add (GObject     *object,
+                              const gchar *description)
+{
+  /* silently ignore objects that are already registered */
+  if (object != NULL
+      && g_hash_table_lookup (objects_table, object) == NULL)
+    {
+      objects_table_count++;
+      g_object_weak_ref (G_OBJECT (object), appfinder_refcount_debug_weak_notify, NULL);
+      g_hash_table_insert (objects_table, object, g_strdup (description));
+    }
+}
+
+
+
+static void
+appfinder_refcount_debug_init (void)
+{
+  objects_table = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, g_free);
+}
+
+
+
+static void
+appfinder_refcount_debug_finalize (void)
+{
+  /* leakup refcount hashtable */
+  APPFINDER_DEBUG ("%d objects leaked, %d registered",
+                   g_hash_table_size (objects_table),
+                   objects_table_count);
+
+  g_hash_table_foreach (objects_table, appfinder_refcount_debug_print, NULL);
+  g_hash_table_destroy (objects_table);
+}
+#endif
+
+
+
 static void
 appfinder_window_destroyed (GtkWidget *window)
 {
@@ -113,6 +180,7 @@ appfinder_window_new (const gchar *startup_id,
   window = g_object_new (XFCE_TYPE_APPFINDER_WINDOW,
                          "startup-id", IS_STRING (startup_id) ? startup_id : NULL,
                          NULL);
+  appfinder_refcount_debug_add (G_OBJECT (window), startup_id);
   xfce_appfinder_window_set_expanded (XFCE_APPFINDER_WINDOW (window), expanded);
   gtk_widget_show (window);
 
@@ -300,6 +368,23 @@ appfinder_daemonize (void)
 
 
 
+static void
+appfinder_signal_handler (gint signum)
+{
+  static gboolean was_triggered = FALSE;
+
+  /* avoid recursing this handler */
+  if (was_triggered)
+    return;
+  was_triggered = TRUE;
+
+  APPFINDER_DEBUG ("received signal %s", g_strsignal (signum));
+
+  gtk_main_quit ();
+}
+
+
+
 static DBusConnection *
 appfinder_dbus_service (const gchar *startup_id)
 {
@@ -396,6 +481,8 @@ main (gint argc, gchar **argv)
   DBusConnection *dbus_connection = NULL;
   const gchar    *startup_id;
   GSList         *windows_destroy;
+  const gint      signums[] = { SIGINT, SIGQUIT, SIGTERM, SIGABRT };
+  guint           i;
 
   /* set translation domain */
   xfce_textdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR, "UTF-8");
@@ -436,6 +523,14 @@ main (gint argc, gchar **argv)
   if (opt_quit)
     return appfinder_dbus_quit ();
 
+  /* setup signal handlers to properly quit the main loop */
+  for (i = 0; i < G_N_ELEMENTS (signums); i++)
+    signal (signums[i], appfinder_signal_handler);
+
+  /* if started with the xfrun4 executable, start in collapsed mode */
+  if (!opt_collapsed && strcmp (*argv, "xfrun4") == 0)
+    opt_collapsed = TRUE;
+
   /* become the serivce owner or ask the current
    * owner to spawn an instance */
   if (G_LIKELY (!opt_disable_server))
@@ -458,6 +553,10 @@ main (gint argc, gchar **argv)
        return EXIT_FAILURE;
     }
 
+#ifdef DEBUG
+  appfinder_refcount_debug_init ();
+#endif
+
   /* create initial window */
   appfinder_window_new (NULL, !opt_collapsed);
 
@@ -465,8 +564,6 @@ main (gint argc, gchar **argv)
 
   gtk_main ();
 
-  xfconf_shutdown ();
-
   /* release the model cache */
   if (model_cache != NULL)
     g_object_unref (G_OBJECT (model_cache));
@@ -479,12 +576,20 @@ main (gint argc, gchar **argv)
 
   if (windows != NULL)
     {
+      /* avoid calling appfinder_window_destroyed */
       windows_destroy = windows;
       windows = NULL;
 
-      /* destroy all windows without poking gtk_main_quit */
+      /* destroy all windows */
       g_slist_foreach (windows_destroy, (GFunc) gtk_widget_destroy, NULL);
+      g_slist_free (windows_destroy);
     }
 
+  xfconf_shutdown ();
+
+#ifdef DEBUG
+  appfinder_refcount_debug_finalize ();
+#endif
+
   return EXIT_SUCCESS;
 }
